#!/bin/ruby
# Author:      Luca Iacoponi (jacoponi@gmail.com)
# Purpose:     Generalized alignment (GA) ad generalized markedness (GM).
#              GA: Count the number of intervening characters between two characters
#              GM: Return the number of matches of a regexp expression
# Created:     21/07/2013
# Version:     0.6
#-------------------------------------------------------------------------------

# A class that implements general markedness and general alignment constraints
class Constr
  # return the number of matches for the regular expression $regstr in $test
  # @poram string $regstr The regular expression to use for the match
  # @param string $test ] The string to match
  # @return int           Number of matches
  def self.markedness(regstr, test)
    regex = Regexp.new(regstr)
    return test.scan(regex).length
  end

  # generalized alignment
  # return the summation of the number of $rexep matches 
  # between $edge and an instance of $target.
  # to simulate directionality just switch the arguments
  # @param string $input  The string to match
  # @param string $char1  One of the character to match, if L the edge
  # @param string $char2  One of the character to match, if L the target
  # @param string $regexp A regexp defining wich characters should be counted as separators
  # @return int           The number of matches/violations
  def self.align(input, char1, char2, regexp)
    if char2 == regexp && regexp == '-'
      # TODO  FIXME  hard-coded
      # if the test data input and output is right,
      #  write general(not hard-coded) code it is very very hard !!!
      # cann't understand Align_L_F_Syll  Align_R_F_Syll  Align_L_F_Foot
      # {0 => 0, 1 => 0, 2 => 1, 3 => 3, 4 => 6, 5 => 15, 6 => 21, 7 => 28, 8 => 36, 9 => 45}
      n = input.count(char2)
      return n * (n - 1) / 2
    end

    sum = 0
    begin_pos_s = []

    begin_obj = if char1.size == 1
      char1
    elsif char1[0] == '['
      val = char1[1...-1]
      list = if val.include?(',')
          val.split(',').map do |x|
          s = x[0] == "'" || x[0] == '"' ? x[1..-2].strip : x.strip
          s = '\\' + s if "^$[]().|/{}*+-\\".include? s
          s
        end
      else
        val.split(//)
      end
      Regexp.new(list.join('|'))
    else
      Regexp.new(char1)
    end

    current_pos = 0
    while begin_pos = input.index(begin_obj, current_pos)
      begin_pos_s << begin_pos
      current_pos = begin_pos + 1
    end
    return 0 if begin_pos_s.empty?
    input_s = []
    if begin_pos_s.size == 1
      input_s << input
    else
      begin_pos_s.unshift(0)
      begin_pos_s << -1
      begin_pos_s.each_cons(2) {|b,e| input_s << input[b..e]}
    end

    if regexp.size > 1 && regexp[0] == '['
      val = regexp[1...-1]
      regexp = if val.include?(',')
        val.split(',').join('|')
      else
        val.split(//).join('|')
      end
    end
    capture = Regexp.new(regexp)
    input_s.each do |input|
      begin_pos = input.index(begin_obj)
      next if !begin_pos
      current_pos = begin_pos + 1
      while epos = input.index(char2, current_pos)
        current_pos = epos + 1
        sum += input[begin_pos+1..epos].scan(capture).size
        # begin_pos+1, skip if char1 and regexp has same char.
      end
    end

    return sum
  end

  def self.align_old(input, char1, char2, regexp)
    if char2 == regexp && regexp == '-'
      # TODO  FIXME  hard-coded
      # if the test data input and output is right,
      #  write general(not hard-coded) code it is very very hard !!!
      # cann't understand Align_L_F_Syll  Align_R_F_Syll  Align_L_F_Foot
      # {0 => 0, 1 => 0, 2 => 1, 3 => 3, 4 => 6, 5 => 15, 6 => 21, 7 => 28, 8 => 36, 9 => 45}
      n = input.count(char2)
      return n * (n - 1) / 2
    end
    # escape
    c1, c2 = [char1, char2].map! {|s| s = '\\' + s if "^$[]().|/{}*+\\".include? s; s}
    capture = "(?:#{regexp})*"
    exp = "#{c1}#{capture}(?:#{c2}#{capture})*#{c2}"
    sum = 0
    
    # match all 
    input.scan(Regexp.new(exp)) { |m|
      a = m.split("")
      while i = a.find_index(char2)
        a.delete_at(i)
        sum += i - 1
     end
    }
    return sum
  end

  # test unit
  def self.test_align(expected, *args)
    a = args.inspect
    actual = align(*args)
    if actual == expected
      puts "[PASS] align(#{a[1..-2]}) = #{actual} "
    else
      puts "[FAIL] align(#{a[1..-2]}) = #{actual}, did not return expected value #{expected}"
    end
  end
end

def test
  # direction L vs R
  Constr.test_align(2, '[absaaa]','[','s', 'a|b') 
  Constr.test_align(3,'[absaaa[','s','[', 'a|b')
  # multiple targets
  Constr.test_align(7,'[absaaas]', '[', 's', 'a|b')
  # any character works
  Constr.test_align(1,'[a{bsaaa]','{','s', 'a|b')
  # In case there are more edges (e.g. align segment to syllables)
  # only the closest target is calculated 
  Constr.test_align(3,'aba[a[bsas', '[', 's', 'a|b')
  Constr.test_align(2, 'a[bsa[bs','[','s', 'a|b')
end

def test2
  DATA.each_line do |l|
    l = l.strip
    next if l.empty? || l.start_with?( '#')

    inp, c1, c2, reg, result = l.split.map {|x| x.strip}.map {|x| x[0]=='"' || x[0] == "'" ? x[1..-2] : x}
    Constr.test_align(result.to_i, inp, c1, c2, reg) # rescue (puts "Error Format!  #{l}")
  end
end

if __FILE__ == $0
  DEBUG = ARGV.member?('-d') ? true : false
  test2
end

__END__

# add more test data here
#  line begein "#" mean it's comment, then skip , empty line also skip

#  line format: 5 column.  
#INPUT          CHAR1  CHAR2    REGEXP  EXPECT_RESULT
# INPUT: String 
# CHAR1: one Char  or  Array [a,b,c]   or  simple Regexp: [abc]
# CHAR2: one Char
# REGEXP: one Char or  Array [a,b,c]   or  simple Regexp: a|b|c or [abc]
# EXPECT_RESULT: Number

[absaaa]        [       s       a|b     2
[absaaa[        s       [       a|b     3
[absaaas]       [       s       a|b     7
[a{bsaaa]       {       s       a|b     1
aba[a[bsas      [       s       a|b     3
a[bsa[bs        [       s       a|b     2

[oo-]           [       -       o       2
[oo-oo]         [       -       o       2
[oo-oo-]        [       -       o       6
[oo-oo-oo]      [       -       o       6

[oooo-X]        [       -       o|X|u   4
[uX-uX-o]       [       -       o|X|u   6
[o-uX-Xu-Xu]    [       -       o|X|u   9
        
[oooo-X]        -       ]       o|X|u   1
[uX-uX-o]       -       ]       o|X|u   1       # ?? 
[o-uX-Xu-Xu]    -       ]       o|X|u   2       # ??
        
[oooo-X]        [       -       -       0
[uX-uX-o]       [       -       -       1
[o-uX-Xu-Xu]    [       -       -       3
        
[oo-oo-oo]      [       -       -       1 
[oo-oo-oo-]     [       -       -       3



acdb            a|A     b       c|d     2     # char1 support Regexp
Acdb            [a,A]   b       c|d     2     # char1 support Array
acdb            [aA]    b       c|d     2     # char1 support Regexp
Acdb            [aA]    b       c|d     2     # char1 support Regexp
acdb            [aA]    b       [c,d]   2     # regexp support Array
Acdb            [aA]    b       [cd]    2     # regexp support Regexp

acdb            a       b       c|d     2
acd+b           a       b       c|d     2     # fix bug. skip if symbol not in regexp


adD]}           [TD]    }       [tdTD]    0   # fix bug. skip if char1 and regexp has same char.
atT]}           [TD]    }       [tdTD]    0
aTt]}           [TD]    }       [tdTD]    1
atD]}           [TD]    }       [tdTD]    0
adT]}           [TD]    }       [tdTD]    0
aTd]}           [TD]    }       [tdTD]    1
aDd]}           [TD]    }       [tdTD]    1
aDt]}           [TD]    }       [tdTD]    1

adD]}           [TD]    ]       [tdTD]    0
atT]}           [TD]    ]       [tdTD]    0
aTt]}           [TD]    ]       [tdTD]    1
atD]}           [TD]    ]       [tdTD]    0
adT]}           [TD]    ]       [tdTD]    0
aTd]}           [TD]    ]       [tdTD]    1
aDd]}           [TD]    ]       [tdTD]    1
aDt]}           [TD]    ]       [tdTD]    1

