# Author: Naz Merchant (based on Ranking Diagram class by BT
#

# Temporarily turn off (verbose) warnings, to surpress (some of) the warnings
# about the source code for graphviz.
verbose_val = $VERBOSE
$VERBOSE = false
require 'graphviz'
require_relative 'dq'
require_relative 'epo'
$VERBOSE = verbose_val

# This class uses Graphviz to create ranking diagrams for appropriately
# formed comparative tableaux.
# When a new RankingDiagram object is created, a graphviz graph is
# automatically constructed for the provided comparative tableau.
# Image files can then be created by calling, on the RankingDiagram object,
# instance methods for the supported image formats.
# The ruby-graphviz gem presumes that you have the Graphviz software installed
# and accessible (on your search path). Graphviz is open source graph
# visualization software; for more information, and to download Graphviz,
# go to http://www.graphviz.org/.
class EpoDiagram

  def initialize(ep, label="EpoDiagram")
    @epo = ep
    @graph = GraphViz::new(label, "type" => "digraph" )

    @graph_string = "digraph G {\n"                         #this will create a file containing a graph named by constraint
    @graph_string << "\t graph [charset=UTF-8 ranksep=0.3 pad=.1]\n"
    @graph_string << "\t node [fontname=Arial shape=box color=blue]\n"
    @graph_string << "\t edge [dir=forward penwidth=0.5]\n"
    
    create_graph
    @graph_string << "}"
  end

  def create_graph
    set_graph_properties
    create_nodes
    create_edges
  end
  private :create_graph

  def set_graph_properties
    @graph.graph[:charset] = "UTF-8"
    @graph.graph[:orientation] = "portrait"
    @graph.graph[:ranksep] = "0.3"
    @graph.graph[:pad] = ".1"
    @graph.node[:fontname] = "Arial"
    @graph.node[:shape] = "box"           #AP- added
    #@graph.node[:style] = "rounded"       #AP- added
    @graph.node[:color] =  "blue"         #AP- added
    @graph.node[:penwidth] = "0.5"        #AP- added
    @graph.edge[:dir] = "forward"            #AP- gives arrowheads as default
    @graph.edge[:penwidth] = "0.5"        #AP- added

  end
  private :set_graph_properties

  # Create a graph node for each language.  Languages are stored
  # in the EPO as strings, and so the language name is its string value
  def create_nodes
    @epo.langs.each do |l|
      # Add a node for each constraint, using the constraint name
      # as both node name and node label.
      @graph.add_node(l).label = l
    end
  end
  private :create_nodes

  # Create a directed graph edge for each W-L pair appearing in an ERC of
  # the comparative tableau. The directed edges are defined as going from
  # the W (dominating) constraint to the L (dominated) constraint.
  def create_edges
    equiv_count = 0
    @epo.equivs.each do |equiv|
      previous_lang = nil
      equiv_count = equiv_count + 1
     
       @graph_string << "\t{ rank = same; \n"                            #open {rank=same ... section
                                #Probably: begin subgraph with rank=same
      equiv.each do |l|
        if previous_lang == nil
          # Just set this lang to previous and don't create an edge
          previous_lang = l
        else
          # Create an edge and then set the current language 'l' to be the
          # previous language, setting it up for the next edge creation
          
          # Give the equiv class edges a special character
          e1 = @graph.add_edge(l, previous_lang)
          e1[:constraint] = "false" #so the edge doesn't affect node rank assignment
          e1[:color] = "blue:green:blue"
          e1[:arrowhead]="none"
          
          @graph_string << "\t\t\"#{previous_lang}\"->\"#{l}\" [color=\"blue:green:blue\" arrowhead=none];\n"      #note order -AP ... and quotes
          g1 = @graph.get_node(l)
          g2 = @graph.get_node(previous_lang)
          previous_lang = l
        end
            					#end subgraph
      end
      @graph_string << "\t}\n"
      
    end
    
    @epo.pos.each do |po|
      po_edge = @graph.add_edge(po[0], po[1])
      po_edge[:color] = "red"
      @graph_string << "\t\"#{po[0]}\" ->\"#{po[1]}\" [color=red];\n"
    end
  end
  private :create_edges

  # Generate a png (portable network graphics) image of the graph,
  # and store it in a file named as per the parameter.
  #--
  # Ruby-graphviz has trouble with filepaths with spaces in them,
  # so add double-quotes around the filepath/name.
  #++
  def png(outfilename)
    @graph.output(:png => "#{outfilename}.png".dq)
  end
  
  def gv(outfilename)
    @graph.output(:gv => "#{outfilename}.gv".dq)
  end

  # Generate an eps (encapsulated postscript) image of the graph,
  # and store it in a file named as per the parameter.
  #--
  # Ruby-graphviz has trouble with filepaths with spaces in them,
  # so add double-quotes around the filepath/name.
  #++
  def eps(outfilename)
    @graph.output(:eps => "#{outfilename}.eps".dq)
  end

  def get_gv
    return @graph_string
  end
  
end # class epoDiagram

