# Author: Naz Merchant 
#

require_relative 'competition'
require_relative 'competition_list'
require_relative 'comparative_tableau'
require_relative 'derivation'
require_relative 'path'
require_relative 'operation'
require_relative 'ot_utils'
require_relative 'rcd'

# A FactorialTypologyHS2 object determines the set of languages defined
# by a constraint set, an operation set, and a set of inputs. It does so by calculating
# the set of derivations from the inputs for each total order, aggregating those
# total orders that produce the same set of derivations.
#
# Wanted: augment this so that it also calculates the operation vectors defined by a 
# typology and the partial orders imposed on the opvects by the languages in the typo.

class Factorial_Typology_HS_Op
  
  attr_reader :langs, :derivs_h, :paths_h
  
  def initialize(inputs, ops, cons)
    @inputs = inputs
    @ops = ops
    @cons = cons

    @langs = Array.new    
    @derivs_h = Hash.new
    @paths_h = Hash.new
    calculate_factorial_typology
  end
  
  # This method produces a hashtable, one for each total order, consisting of
  # derivations, one for each input. (NNM: numerous mods since this was near accurate) 
  # The hashtable is the language for that total order.
  def calculate_factorial_typology
    
    @inputs.each do |inp|
      paths = Array.new
      @paths_h[inp] = paths
      Path.spawn_paths(inp, paths, @ops)      
    end
    
    create_consistent_path_groups
    calculate_typology_with_derivations
  end
  
  # A derivation is a maximal set of consistent paths
  def create_consistent_path_groups
    @inputs.each do |input|
      @derivs_h[input] = Array.new
      for pos in 0..@paths_h[input].length - 1
        path_recurse(Derivation.new(input), input, pos)
      end
    end
  end
  
  def path_recurse(deriv, input, pos)      
    # Get the path under consideration.     
    path = (@paths_h[input]).at(pos) 
    
    # If current derivation + path are consistent, 
    #             recurse(path_group + path, input, pos+1)   
    if deriv.is_consistent?(path)    
      deriv_n = deriv.dup
      deriv_n.add_path(path)
      p_len = @paths_h[input].length

      if pos < p_len - 1
        for pos_n in pos+1..p_len-1
          path_recurse(deriv_n, input, pos_n)
        end
      else
        @derivs_h[input] << deriv_n if !consistent_deriv(deriv_n, input)
      end
   
    else
    # If path_group + path inconsistent 
    #                      and pos = last
    #                      and no other path_groups have been spawned at this level  
    #                        (this is checkable by determining consistency w/extant path_groups)         
    #             path_group is a final path, unrefinable, and so recordable.
      @derivs_h[input] << deriv if (pos == @paths_h[input].length - 1) && !consistent_deriv(deriv, input)   
    end
  end

  # Is the ranking set of 'deriv' consistent with other recorded derivations?
  def consistent_deriv(deriv, input)
    @derivs_h[input].each do |d|
      consistent = deriv.derivation_consistent?(d)
      return true if consistent
    end
    return false
  end

=begin
  # Two paths for a single input are either ranking-set-identical or disjoint, (not true: 
  # two candidates in a cset can be viol prof identical, spawning their own paths which, later,
  # add differing ranking requirements)
  # so RCD can be deployed to determine if a pair (or more) should be 
  # grouped together in a derivation.
  
  # Warning, this method is not used in the current HS facttyp builder. Remove?
  def consolodate_paths_into_derivations
    @inputs.each do |input|  
      # Initialize @deriv_h to have empty arrays for each input
      @derivs_h[input] = Array.new
      modded = true
      while modded
        modded = false
        @paths_h[input].each do |p|
          #check to see if compat w/deriv in there,
            # if compat, add it to that deriv
            # if not compat, add it to a new deriv
          #puts "P: #{p.to_s}"
          merged = false
          @derivs_h[input].each do |d|
            # Is p compat with d?
            #puts "D: #{d.to_s}"
            ct = d.get_comparative_tableau
            ct.concat(p.get_comparative_tableau)
            rcd = Rcd.new(ct)
            if rcd.consistent?
              if !d.contains_path?(p)
                d.add_path(p)
                modded = true
              end
              merged = true 
            end
          end
          
          if !merged
            new_d = Derivation.new(input)
            new_d.add_path(p)
            @derivs_h[input] << new_d
            modded = true
          end
          
        end #@paths_h loop
      end #while loop
    end #@inputs.each loop
  end
=end

  def calculate_typology_with_derivations
    lang = Hash. new
    construct_product(0, lang)
  end
  
  def construct_product(pos, lang)
    @derivs_h[@inputs[pos]].each do |d|
      ct = d.get_comparative_tableau
      lang.values.each {|deriv| ct.concat(deriv.get_comparative_tableau)}
      rcd = Rcd.new(ct)
      
      if rcd.consistent?
        lang_n = Hash.new
        lang_n[@inputs[pos]] = d
        lang.each_pair {|k, v| lang_n[k] = v}
        pos == @inputs.length-1 ? @langs << lang_n : construct_product(pos+1, lang_n)
      end
    end
  end

  def get_undifferentiated_outputs
    outs = Set.new
    @langs.each do |lang| 
      @inputs.each {|i| lang[i].get_final_optima.each {|o| outs.add(o)}}
    end
     
    return outs
  end
  
  def get_comparative_tableau(lang)     
    ct = Comparative_tableau.new("", @cons)
    lang.values.each {|d| d.get_competition_list.each {|c| ct.add_competition(c)}}
    return ct
  end
  
  # Returns an array of comparative tableau, one for each language in the typology
  def get_languages_as_comparative_tableau
    cts = Array.new
    @langs.each {|d_ht| cts << get_comparative_tableau(d_ht)}
    return cts
  end
 
  def to_s
    # Only display the derivations for testing purposes
    
    # Languages are simply derivations
    s = "Languages: \n"
    @langs.each {|l| s += l.to_s + "\n"}
    return s
  end
  
end
