#!/bin/ruby
# Author:      Luca Iacoponi (jacoponi@gmail.com)
# Purpose:     Given a regular expression, generate all possible string matching
#              the regular expression. Can also generates two lists that simulate
#              OT's input:candidate pairs.
# Created:     21/07/2013
# Version:     0.6
#-------------------------------------------------------------------------------

# TODO
# split the regexp into its operators, and 
# return the cartesian product of the generated expressions
# may improve the performance of the algorithm.

# helper class for Gen (private?)
class Gen_helper
  # all functions in this class are generators,
  # using generators instead of arrays can save memory usage
  # generate all possible combinations of characters in $lexicon of length $n
  def self.gen_comb(n, upto, lexicon)
    chars = lexicon.chars
    if n == 0
      yield ""
    else
      if upto
        # generate a string of length < N
        gen_comb(n-1, true, lexicon) { |s|
          yield s
        }
      end

      # generate string of length N
      gen_comb(n-1, false, lexicon) { |s|
        for i in chars do
          yield s + i.chr
        end
      }
    end
  end

  def self.gen_one(regex, n, upto, lexicon)
    reg = Regexp.new(regex)
    gen_comb(n, upto, lexicon) { |s|
      if reg.match(s)
        yield s
      end
    }
  end
end

# The class Gen
class Gen
  # generate all possible ASCII characters combinations of length $n that match $regex
  # @param string $regex The regular expression
  # @param int $n        Maximum lenght of the string
  # @param boolean $upto If true, generate all strings from length 0 to $n
  #                      if false, only strings of length $n
  # @param string @lexicon The characters to be generated in the combination
  # @return array of string
  def self.gen_one(regex, n, upto, lexicon)
    result = []
    Gen_helper.gen_one(regex, n, upto, lexicon) {|x| result << x}
    return result
  end

  # generate a set of string (input) that matches the regular expression r1
  # for each of the generated string, further generates a set of strings
  # that matches the regular expression r2
  # note: you only need to generate the second list once.
  # @param string $r1  First regular expression generator (input)
  # @param string $r2  Second regular expression generator (candidates)
  # @param int $n(1|2) Max string length of input and candidates
  # @param boolean @upto = if true, generate all strings from length 0 to $n
  #                        if false, only strings of length $n
  # @param string lexicon The characters to be generated in the combination
  # @return array of string
  def self.gen_two(r1, r2, n1, n2, upto1, upto2, lexicon1, lexicon2, limitIOsamelength)
    result = []
    result_g2 = gen_one(r2, n2, upto2, lexicon2)
    Gen_helper.gen_one(r1, n2, upto2, lexicon2) {|x|
        if limitIOsamelength
          r2 = result_g2.select {|x1| x1.size == x.size}
          if !r2.empty?
            result << x
            result << r2
          end
        else
          result << x
          result << result_g2
        end
    }
    return result
  end

  # test output
  def self.test_and_output(f, *args)
    puts "#{f.name}(#{args}) = #{f.call(*args).inspect}"
  end

  # test unit 
  def self.test
    tregexp1 = 'C*V+C*'
    tregexp2 = "C*V+C"
    lexicon1 = 'CV'
    lexicon2 = 'CV'
    upto1 = true
    upto2 = true
    n2 = 3
    n1 = 4
    test_and_output(Gen.method(:gen_one), tregexp1, n1, false, lexicon1)
    test_and_output(Gen.method(:gen_one), tregexp1, n1, true, lexicon1)
    test_and_output(Gen.method(:gen_two), tregexp1, tregexp2, n1, n2, upto1, upto2, lexicon1, lexicon2, false)
    test_and_output(Gen.method(:gen_two), tregexp1, tregexp2, n1, n2, upto1, upto2, lexicon1, lexicon2, true)
  end 
end


if __FILE__ == $0
  # run the test
  Gen.test()
end
