# Author: Naz Merchant
#

# A moat is a collection of epos, one for each constraint



require_relative 'epo.rb'
require_relative 'constraint.rb'
require_relative 'candidate.rb'
require_relative 'competition.rb'

require 'set'

#require 'epo.rb'
#require 'constraint.rb'
#require 'candidate.rb'
#require 'competition.rb'



class Moat

  attr_reader :constraints
  attr_reader :languages
  attr_reader :epos

  # A moat requires a VT (currently unimplemented) and a set of constraints (this
  # is probably removable after the VT portion is completed).  The constructor
  # creates a hash table containing an EPO for each constraint.
  # TO DO: construct the Epos based on the VT -- currently its built from a
  # set of languages

  # The Moat initialization method receives a single competition and assumes that
  # each of the candidates in the competition is a possible winner.
  # Basically, it's assuming that the Mink Sum has been applied and that
  # harmonically bound candidates have been removed.

  # Eventually the languages parameter of initialize should be removed (as should the constraint
  # parameter for similar reasons).  Languages should be constructed from the competition,
  # with each candidate corresponding to a single language.

  def initialize(competition)
    #@constraints = constraints
    #@languages = languages

    @constraints = competition.constraint_list
    @languages = Hash.new
    lang_num = 0
    competition.each {|comp| @languages[comp.output] = comp}
    @label = "Moat"
    @orders = create_total_orders
    create_epos
  end

  # Note: this method has been copied to ot_utils.rb and should be 
  # removed from moat.rb
  
  def create_total_orders
    ords = continue_sequence(@constraints)
    return ords
  end

  def continue_sequence(cons)
    continue_set = Set.new
    cons.each do |c|
      if cons.length > 1
        cons_wo_c = Array.new(cons)
        cons_wo_c.delete(c)

        sets = continue_sequence(cons_wo_c)
        sets.each do |seq|
          order = Array.new
          order << c
          order.concat(seq)
          continue_set.add(order)
        end
      else
        # Length here should be 1, meaning that we've reached the end of the
        # constraint set, and so a new order should be created with c in it.
        order = Array.new
        order << c
        continue_set.add(order)
      end
    end

    return continue_set
  end

  def create_epos
    @epos = Hash.new
    @constraints.each {|c| @epos[c] = Epo.new(@languages.keys)}
    @orders.each do |ord|
      curr_langs = @languages.keys
      ord.each do |c|
        # Equivalence first

        # get all langs that min violate c
        min_langs = get_min_langs(c, curr_langs)

        # set these equiv on c, if multiple langs are equiv, i.e. if not a DIS yet
        @epos[c].set_equiv(min_langs) if min_langs.size > 1

=begin
        # Partial order next: want min_langs < current_lang - min_lang
        excluded_langs = curr_langs - min_langs

        min_langs.each do |cl|
          excluded_langs.each do |el|
            @epos[c].set_po(cl, el)
          end
        end
=end
        # Partial orders next
        # First, get the next order on the oriented (by constraint c) permutahedron

        # Check to see if c is in the last position -- do nothing if so
        c_pos = ord.index(c)
        if c_pos < ord.length - 1
          next_ord = Array.new(ord)
          next_ord.insert(c_pos + 2, c)
          next_ord.delete_at(c_pos)

          ord_lang = get_lang(ord)
          next_lang = get_lang(next_ord)
          if ord_lang != next_lang
            @epos[c].set_po(ord_lang, next_lang)
          end
        end

        # remove all other lang
        curr_langs = min_langs
      end
    end
    @epos.values.each {|epo| epo.remove_dup_pos}
  end

  def get_lang(order)
    langs = Array.new(@languages.keys)
    order.each do |c|
      min_langs = get_min_langs(c, langs)
      langs = min_langs
    end

    return langs[0]
  end

  # Receives a constraint a set of language names (the keys in @language)
  # and returns a vector containing the names of languages (from those passed in)
  # that minimally violate the constraint.
  def get_min_langs(constraint, langs)

    # First, get the minimum constraint violation for all the languages on this constraint
    min = -1
    langs.each do |lang|
      lang_val = @languages[lang].get_viols(constraint)
      if min == -1
        min = lang_val
      else
        min = lang_val if lang_val < min
      end
    end

    # Second, put all those languages with that violation in a vector
    min_langs = Array.new
    langs.each do |lang|
      lang_val = @languages[lang].get_viols(constraint)
      min_langs << lang if lang_val == min
    end

    return min_langs
  end

  def to_s
    moat_s = String.new
    moat_s << @label + "\n"
    @epos.keys.each {|con| moat_s << con.name + ": " + @epos[con].to_s}
    return moat_s
  end

  def ords_to_s
    ords_s = "Orders: \n" + @orders.length.to_s + "\n"
    @orders.each do |ord|
      ord.each do |c|
        ords_s = ords_s + c.name + " "
      end
      ords_s = ords_s + "\n"
    end
    return ords_s
  end

  def Moat.get_test_moat()
    c1 = Constraint.new("c1", "c1")
    c2 = Constraint.new("c2", "c2")
    c3 = Constraint.new("c3", "c3")

    constraints = Array.new
    constraints << c1
    constraints << c2
    constraints << c3

    # Need candidates and competitions
    cand1 = Candidate.new("in1", "o1", true, constraints)
    cand2 = Candidate.new("in1", "o2", true, constraints)
    cand3 = Candidate.new("in1", "o3", true, constraints)

    cand1.set_viols(c1, 0)
    cand1.set_viols(c2, 0)
    cand1.set_viols(c3, 1)

    cand2.set_viols(c1, 0)
    cand2.set_viols(c2, 1)
    cand2.set_viols(c3, 0)

    cand3.set_viols(c1, 1)
    cand3.set_viols(c2, 0)
    cand3.set_viols(c3, 0)

    comp = Competition.new
    comp << cand1
    comp << cand2
    comp << cand3

    test_moat = Moat.new(comp)
    return test_moat
  end

  def Moat.get_lombardi_moat()

    c1 = Constraint.new("x_Voi", "x_voi")
    c2 = Constraint.new("y_Agr", "y_Agr")
    c3 = Constraint.new("z_FV", "z_FV")
    c4 = Constraint.new("w_FHV", "w_FHV")

    constraints = Array.new
    constraints << c1
    constraints << c2
    constraints << c3
    constraints << c4

    # Need candidates and competitions
    cand1 = Candidate.new("in1", "o1", true, constraints)
    cand2 = Candidate.new("in1", "o2", true, constraints)
    cand3 = Candidate.new("in1", "o3", true, constraints)
    cand4 = Candidate.new("in1", "o4", true, constraints)
    cand5 = Candidate.new("in1", "o5", true, constraints)
    cand6 = Candidate.new("in1", "o6", true, constraints)
    cand7 = Candidate.new("in1", "o7", true, constraints)

    cand1.set_viols(c1, 4)
    cand1.set_viols(c2, 2)
    cand1.set_viols(c3, 0)
    cand1.set_viols(c4, 0)

    cand2.set_viols(c1, 2)
    cand2.set_viols(c2, 0)
    cand2.set_viols(c3, 2)
    cand2.set_viols(c4, 2)

    cand3.set_viols(c1, 7)
    cand3.set_viols(c2, 0)
    cand3.set_viols(c3, 3)
    cand3.set_viols(c4, 0)

    cand4.set_viols(c1, 4)
    cand4.set_viols(c2, 0)
    cand4.set_viols(c3, 2)
    cand4.set_viols(c4, 1)

    cand5.set_viols(c1, 3)
    cand5.set_viols(c2, 2)
    cand5.set_viols(c3, 1)
    cand5.set_viols(c4, 0)

    cand6.set_viols(c1, 6)
    cand6.set_viols(c2, 0)
    cand6.set_viols(c3, 4)
    cand6.set_viols(c4, 0)

    cand7.set_viols(c1, 0)
    cand7.set_viols(c2, 0)
    cand7.set_viols(c3, 4)
    cand7.set_viols(c4, 3)


    comp = Competition.new
    comp << cand1
    comp << cand2
    comp << cand3
    comp << cand4
    comp << cand5
    comp << cand6
    comp << cand7

    test_moat = Moat.new(comp)
    return test_moat

  end
end