# Author: Nazarre Merchant
#

# An OpVec has a winning candidate, a base candidate (which may be the same 
# as the winner), an operation its associated with, and an integer value
# for each constraint, calculated by subtracting the winner's val from the base's.

require_relative 'candidate_hs.rb'

class OpVec
  attr_reader :winner, :base, :operation, :constraints

  def initialize(winner, base, operation, constraints)
    @winner = winner
    @base = base
    @operation = operation
    @constraints = constraints
  end

  def constraint_list
    @constraints
  end

  # Returns the value of base minus winner in the constraint con dimension
  def get_val(con)
    @winner.get_viols(con) - @base.get_viols(con)
  end

  # Two opvecs are equal if they have the same vals in each constraint and 
  # were produced by the same operation -- winner and base are left unchecked.
  def ==(other)
    return false unless @operation==other.operation
    @constraints.each {|c| return false if get_val(c) != other.get_val(c)}
    true
  end

  # The same as ==
  def eql?(other)
    self==other
  end
  
  def unique_name
    name = "#{@operation.name}(#{winner.output}) "
    @constraints.each {|c| name += ":#{get_val(c)}" }
    return name
  end

  def short_name
    sn = "#{@operation.name}.#{@winner.output}"
    return sn
  end

  # Represent the candidate with a string.
  # * The candidate's label.
  # * The input and output strings, separated by " --> "
  # * A list of constraints and the number of violations of each.
  #   If a constraint hasn't been assigned a violation count, display '?'.
  # * The remark, if any.
  # * If this is a merged candidate, the outputs of the individual
  #   merge candidates are listed, one per line.
  def to_s
    val_s = " "
    @constraints.each do |c|
      val_c = get_val(c)
      val_s += " #{c}:#{val_c}"
    end
    if @remark then remark_s = "  #{@remark}" else remark_s = "" end
    win_base_s = "#{@winner.input.to_s} --> #{@winner.output.to_s} ~ #{@base.output.to_s}"
    full_s = "#{win_base_s}: #{val_s}"
    return full_s
  end

end 
