# Author: Naz Merchant
#

require_relative 'competition.rb'
require_relative 'competition_list.rb'
require_relative 'comparative_tableau.rb'
require_relative 'hierarchy.rb'

# A path is an ordered set of competitions.  Each competition
# has exactly one candidate marked as optimal (even if co-optima exist). The ordering on
# the competitions is the obvious one: the winning candidate's output should match the next 
# competition's input.
#
# Currently there is no checked that this is a viable path, amongst other things 
# that could go wrong there is no checking that there is a viable ranking that maps input to output.

class Path < Array

    def to_s
      return "Empty" if self.empty?
      s = self[0][0].input
      # Warning: each competition added to path should have exactly one winner 
      self.each {|comp| s = "#{s} -> #{(comp.winners)[0].output}"}

      return s
    end
    
    def get_final_optima
      self[self.length-1].winners[0].output
    end
    
    def get_competition_list
      comp_list = Competition_list.new
      self.each{|c| comp_list << c}
      return comp_list
    end
    
    def get_comparative_tableau
      comp_list = get_competition_list
      ct = Comparative_tableau.new("", comp_list.constraint_list)
      ct.add_competition_list(comp_list)
      return ct
    end
    
    # This method creates all paths for a given input consistent with cset. Each candidate in each
    # cset (except HB'ed, DB'ed cands, and prohibited faithful cands) produces a new path. 
    def self.spawn_paths(input, paths, ops, hier=nil)
      self.spawn_paths_r(input, Path.new, paths, ops, hier)
    end
      
    def self.spawn_paths_r(input, path, paths, ops, hier)
      cset = Competition.new
      #ops.each{|op| op.create_candidates(input).each{|cand| cset << cand} }
      ops.each do |op|
        op.create_candidates(input).each{|cand| cset << cand} 
      end
      cset = self.remove_hb_and_db(cset, path, hier)
      if cset.length > 0
        cset.each do |c|      
          # Only spawn paths for those that could win on hier
          if hier == nil || hier.winners(cset).include?(c)
            c.opt=true
            path_n = path.dup
            path_n << cset.dup
            # If c is the faithful candiate then convergence achieved, otherwise spawn new paths
            c.is_faithful? ? paths << path_n : spawn_paths_r(c.output_as_input, path_n, paths, ops, hier)
            c.opt=false
          end
        end
      else
        # There are no outputs for this input (a possibility in some HS systems),
        # so we have converged by default and have reached the end of path.
        paths << path
      end
    end
    
    def self.remove_hb_and_db(cset, path, hier)
      cset_not_hb = Competition.new
      cset_new_outs = Competition.new
      
      #cset = remove_comparative_struct_lacking(cset)
      
      cset.each {|c| cset_new_outs << c if self.new_output?(c.output_as_input, path)}
      
      cset_new_outs.each do |cand|
        ct = Comparative_tableau.new("check_with_rcd", cand.constraint_list)
        winner_comp = self.construct_competition_with_winner(cand, cset_new_outs)
        # check to see if winner is possibly optimal
        ct.add_competition(winner_comp)
        path.get_competition_list.each {|c| ct.add_competition(c)}
        rcd_result = Rcd.new(ct)
        # If the CT is inconsistent, then the candidate is harmonically bound
        #cset_nhb << cand if rcd_result.consistent? && new_output?(cand.output_as_input, path)
        cset_not_hb << cand if rcd_result.consistent?
      end
      
      return cset_not_hb
    end
    
    # method below is code taken from BT's factorial typology (now defunct)
  
    # Returns a copy of the competition _competition_, altered so that the
    # candidate <em>candidate_winner</em> is the winner.
    # A duplicate is made of each of the candidates so that the altered
    # opt field doesn't affect objects referenced elsewhere.
    def self.construct_competition_with_winner(candidate_winner, competition)
      winner = candidate_winner.dup
      losers = competition.reject{|c| c==candidate_winner}
      losers = losers.map{|c| c.dup}
      winner.opt=true; losers.each{|c| c.opt=false}
      new_comp = Competition.new
      new_comp.push(winner)
      losers.each{|loser| new_comp.push(loser)}
      return new_comp
    end
    
    def remove_comparative_struct_lacking(cset)
      return cset
    end
    
    # Check to see if this output has appeared as an input in previous stages of derivation.
    # NB: not guaranteed convergence (ala Moreton) because this system permits candidate sets
    # without the faithful candidate (see HS PAKA for example where omission is nec.)
    def self.new_output?(output, path)
     # []
      path.get_competition_list.each do |comp|
        return false if comp[0].input == output 
      end
      return true
    end
 
end