# Author: Bruce Tesar
# 

# Temporarily turn off (verbose) warnings, to surpress (some of) the warnings
# about the source code for graphviz.
verbose_val = $VERBOSE
$VERBOSE = false
require 'graphviz'
require_relative 'dq'
$VERBOSE = verbose_val

# This class uses Graphviz to create ranking diagrams for appropriately
# formed comparative tableaux.
# When a new RankingDiagram object is created, a graphviz graph is
# automatically constructed for the provided comparative tableau.
# Image files can then be created by calling, on the RankingDiagram object,
# instance methods for the supported image formats.
# The ruby-graphviz gem presumes that you have the Graphviz software installed
# and accessible (on your search path). Graphviz is open source graph
# visualization software; for more information, and to download Graphviz,
# go to http://www.graphviz.org/.
class RankingDiagram

  def initialize(ct, label="RankingDiagram")
    @comp_tableau = ct
    @graph = GraphViz::new(label, "type" => "digraph" )
    create_graph
  end

  def create_graph
    set_graph_properties
    create_nodes
    create_edges
  end
  private :create_graph

  def set_graph_properties
    @graph.graph[:charset] = "UTF-8"
    @graph.graph[:orientation] = "portrait"
    @graph.graph[:ranksep] = "0.3"
    @graph.graph[:pad] = ".1"
    @graph.node[:fontname] = "Arial"
    @graph.node[:shape] = "box"           #AP- added
    @graph.node[:style] = "rounded"       #AP- added
    @graph.node[:color] =  "blue"         #AP- added  
    @graph.node[:penwidth] = "0.5"        #AP- added    
    @graph.edge[:dir] = "none"            #AP- decommented out
    @graph.edge[:penwidth] = "0.5"        #AP- added
    
  end
  private :set_graph_properties

  # Create a graph node for each constraint, using the name of
  # the constraint as both the name and the label of the node.
  def create_nodes
    @comp_tableau.constraint_list.each do |n|
      # Add a node for each constraint, using the constraint name
      # as both node name and node label.
      @graph.add_node(n.name).label = n.name
    end
  end
  private :create_nodes

  # Create a directed graph edge for each W-L pair appearing in an ERC of
  # the comparative tableau. The directed edges are defined as going from
  # the W (dominating) constraint to the L (dominated) constraint.
  def create_edges
    edge_list = []
    @comp_tableau.each do |erc|
      # Convert sets of W and L constraints to arrays
      w_cons, l_cons = erc.w_cons.to_a, erc.l_cons.to_a
      # Take cartesian product of W and L constraints (W constraint first
      # in the pair), and add the pairs to the edge list.
      edge_list += w_cons.product(l_cons)
    end
    # For each W-L pair in the edge_list, add an edge to the graph.
    edge_list.each { |e| @graph.add_edge(e[0].name, e[1].name) }
  end
  private :create_edges


  def get_depth(c_name)
    doms = count_doms(c_name)
    doms = -1*doms unless (is_dominator(c_name))
    return doms
  end
  
  def is_dominator(c_name)
    @graph.each_edge() do |e|
      return true if (e.node_one == c_name)
    end
    return false
  end
  
  def count_doms(c_name)
    doms = 0
    @graph.each_edge() do |e|  
      doms = doms+1+count_doms(e.node_one) if (e.node_two == c_name)
    end
    return doms
  end


  # Generate a png (portable network graphics) image of the graph,
  # and store it in a file named as per the parameter.
  #--
  # Ruby-graphviz has trouble with filepaths with spaces in them,
  # so add double-quotes around the filepath/name.
  #++
  def png(outfilename)
    @graph.output(:png => "#{outfilename}.png".dq)
  end

  # Generate an eps (encapsulated postscript) image of the graph,
  # and store it in a file named as per the parameter.
  #--
  # Ruby-graphviz has trouble with filepaths with spaces in them,
  # so add double-quotes around the filepath/name.
  #++
  def eps(outfilename)
    @graph.output(:eps => "#{outfilename}.eps".dq)
  end

end # class RankingDiagram

