# Author: Naz Merchant
#

# A derivation is an input and a set of paths.

require_relative 'path.rb'
require_relative 'operation.rb'
require_relative 'hierarchy.rb'
require_relative 'competition.rb'
require_relative 'competition_list.rb'
require_relative 'comparative_tableau.rb'
require_relative 'opvec.rb'

class Derivation

    attr_reader :input, :paths

    def initialize(input, hier, ops)
      @input = input
      @hier = hier
      @ops = ops
      @paths = Array.new
      
      p = Path.new
      @paths << p
      derive_paths(p, @input)
    end

    def derive_paths(path, input)
      cand_set = Competition.new
      @ops.each{|op| op.create_candidates(input).each{|cand| cand_set << cand} }
      return if cand_set.length == 0
      
      winners = @hier.winners(cand_set)
      orig_path = path.dup
      new_path = nil
      
      winners.each do |w|
        if new_path == nil
          new_path = path
        else
          new_path = orig_path.dup
          @paths << new_path
        end
        w.opt=true
        new_path << cand_set.dup
        #So that there is only one optimum for cand sets in paths
        w.opt=false
        derive_paths(new_path, w.output_as_input) unless w.is_faithful?
      end  
    end
     
    def get_competition_list
      comp_list = Competition_list.new
      @paths.each {|p| p.each{|c| comp_list << c}}
      return comp_list
    end
    
    def get_opvecs
      vecs = Array.new
      get_competition_list.each do |comp|
        base = get_base(comp)
        #NB: this doesn't produce the order the derivation imposes on the opvecs
        comp.each {|cand| vecs << OpVec.new(cand, base, cand.operation, cand.constraint_list) }
      end
      return vecs
    end
    
    def get_base(competition)
      base = nil
      competition.each {|cand| base = cand if cand.base?}
      # Arb. select first candidate for the base if none are marked (faith cand is default marked)
      base = competition[0] if base.nil?
      base.base=true
      return base
    end
    
    def get_comparative_tableau
      comp_list = get_competition_list
      ct = Comparative_tableau.new("", comp_list.constraint_list)
      ct.add_competition_list(comp_list)
      return ct
    end

    # Returns a set of outputs that are the final outputs of each path
    def get_final_optima
      outs = Array.new
      @paths.each {|p| outs << p.get_final_optima}
      return outs
    end

    def contains_path?(path)
      @paths.each {|p| return true if p == path} 
      false
    end

    # Two derivations are equivalent if their paths are the same
    def ==(other)
      return false if @input != other.input
      return false if @paths.length != other.paths.length
      @paths.each {|p| return false unless other.contains_path?(p)}
      return true
    end
  
    # The same as ==
    def eql?(other)
      self == other
    end

    def to_s
      s = String.new
      s = @hier.to_s + "\n"
      @paths.each {|p| s = "#{s}\t#{p.to_s}\n"}
      return s
    end

end
