# Author: Naz Merchant
#

# An Epo has a set of languages and two relations on those languages, an
# equivalence relation and a partial order.  These two relations must respect
# the other's relations, so that if a~b, then a must not be less than or greater
# than b, and if a < b, a must not be equivalent to b.

require 'set'

class Epo

  attr_reader :langs
  attr_reader :equivs
  attr_reader :pos

  # Initially each language is placed in its own equivalence relation and no
  # two languages are comparable under the partial ordering relation.
  #
  # The partial ordering relation only keeps immediate dominance relations.
  # The transitive closure of the PO is left out.
  def initialize(languages)
    @langs = languages
    @equivs = Set.new
    @pos = Set.new
    init_equiv_relation
  end

  def init_equiv_relation
    #This method puts each language in its own equivalence class
    @langs.each {|lang| @equivs << Set.new.add(lang)}
  end
  private :init_equiv_relation

  # Receives a set of languages, removes them from their current equiv classes
  # and adds them back in as an equivalent set.  Note that this method always
  # maintains an equivalence relation on the set of languages (assuming that
  # they already are in an equivalence relation).
  def set_equiv(languages)
    agg_langs = Array.new
    languages.each do |lang|
      @equivs.each do |e|
        if e.include?(lang)
          e.each {|l| agg_langs << l}
        end
      end
    end
   
    set_equiv_aggregated(agg_langs.uniq)
  end
  
  # This assumes that 'languages' includes all elements of each equiv class
  # that any element of 'languages' contains
  def set_equiv_aggregated(languages)
    languages.each do |lang|
      # First remove lang from its current equivalence class & delete the
      # equivalence class if it contains no members
      @equivs.each do |e|
        e.delete(lang)
      end

    end
    @equivs << languages
    remove_empty_sets()
  end
  # This is a horrible hack that is currently needed because the set object does not
  # delete objects correctly.
  def remove_empty_sets
    non_empty_equivs = Set.new
    @equivs.each {|e| non_empty_equivs << e if e.size() > 0}
    @equivs = non_empty_equivs
  end
  private :remove_empty_sets

  # Adds to the set @pos an array containing the two params, the first element
  # in the array is the greater-than lang (having fewer viols), the second, the less-than
  # (having more viols).
  # Currently this method has no checking for PO-ness, e.g. cycles could be
  # introduced; also, there is no checking for EPO-ness -- this method could
  # add POs that don't respect the equivalence relations.
  def set_po (great_lang, less_lang)
    @pos << [great_lang, less_lang]
  end


  # Returns all languages that lang dominates, stored in an array
  def get_dominates(lang)
    children = Array.new
    @pos.each do |po|
      children << po[1] if po[0] == lang
    end
    ancestors = Array.new(children)
    children.each do |child_lang|
      ancestors = ancestors + get_dominates(child_lang)
    end
    return ancestors
  end

  def remove_dup_pos
    #puts "REMOVING"
    @langs.each do |lang|
      children = []
      grand_c = []
      # Get immediate children
      @pos.each do |po|
        if po[0] == lang
          children << po[1]
          grand_c = grand_c + get_dominates(po[1])
        end
      end

      # If a non-immediate child = immediate child, delete imm_child po
      children.each do |child|
        #puts "testing #{child}"
        if grand_c.include?(child)
          @pos.delete([lang, child])
          #puts "REMOVED: #{lang} #{child}"
        end
      end

    end
  end

  def to_s
    epo_string = String.new
    @langs.each {|lang| epo_string << lang.to_s + " "}
    epo_string << "\n"
    @equivs.each do |e|
      epo_string << "{"
      e.each {|lang| epo_string << lang.to_s + " "}
      epo_string << "}\n"
    end

    #epo_string << "\n"
    @pos.each do |po|
      epo_string << po[0] + " > " + po[1]
      epo_string << "\n"
    end

    return epo_string
  end

end