# Author: Nazarre Merchant (based on Ranking Diagram class by BT)
#

# Temporarily turn off (verbose) warnings, to surpress (some of) the warnings
# about the source code for graphviz.
verbose_val = $VERBOSE
$VERBOSE = false

require 'graphviz'
require_relative 'dq'
require_relative 'opvec_po'
require_relative 'opvec'

$VERBOSE = verbose_val

# This class uses Graphviz to create ranking diagrams for appropriately
# formed comparative tableaux.
# When a new RankingDiagram object is created, a graphviz graph is
# automatically constructed for the provided comparative tableau.
# Image files can then be created by calling, on the RankingDiagram object,
# instance methods for the supported image formats.
# The ruby-graphviz gem presumes that you have the Graphviz software installed
# and accessible (on your search path). Graphviz is open source graph
# visualization software; for more information, and to download Graphviz,
# go to http://www.graphviz.org/.
class OpvecPODiagram

  def initialize(opv_po, label="OpvecDiagram")
    @po = opv_po
    @graph = GraphViz::new(label, "type" => "digraph" )

    @graph_string = "digraph G {\n"                         #this will create a file containing a graph named by opvecs
    @graph_string << "\t graph [charset=UTF-8 ranksep=0.3 pad=.1]\n"
    @graph_string << "\t node [fontname=Arial shape=box color=blue]\n"
    @graph_string << "\t edge [dir=forward penwidth=0.5]\n"
    
    create_graph
    @graph_string << "}"
  end

  def create_graph
    set_graph_properties
    create_nodes
    create_edges
  end
  private :create_graph

  def set_graph_properties
    @graph.graph[:charset] = "UTF-8"
    @graph.graph[:orientation] = "portrait"
    @graph.graph[:ranksep] = "0.3"
    @graph.graph[:pad] = ".1"
    @graph.node[:fontname] = "Arial"
    @graph.node[:shape] = "box"           #AP- added
    #@graph.node[:style] = "rounded"       #AP- added
    @graph.node[:color] =  "blue"         #AP- added
    @graph.node[:penwidth] = "0.5"        #AP- added
    @graph.edge[:dir] = "forward"            #AP- gives arrowheads as default
    @graph.edge[:penwidth] = "0.5"        #AP- added

  end
  private :set_graph_properties

  # Create a graph node for each opvec.  
  def create_nodes
    @po.opvecs.each do |ov|
      # Add a node for each opvec, must be stored as a string
      @graph.add_node(ov.unique_name).label = ov.unique_name unless @po.omit(ov)
    end
  end
  private :create_nodes

  def create_edges
    @po.pos.each do |po|
      po_edge = @graph.add_edge(po[0].unique_name, po[1].unique_name)
      po_edge[:color] = "red"
      @graph_string << "\t\"#{po[0].unique_name}\" ->\"#{po[1].unique_name}\" [color=red];\n"
    end
  end
  private :create_edges

  # Generate a png (portable network graphics) image of the graph,
  # and store it in a file named as per the parameter.
  #--
  # Ruby-graphviz has trouble with filepaths with spaces in them,
  # so add double-quotes around the filepath/name.
  #++
  def png(outfilename)
    @graph.output(:png => "#{outfilename}.png".dq)
  end
  
  def gv(outfilename)
    @graph.output(:gv => "#{outfilename}.gv".dq)
  end

  # Generate an eps (encapsulated postscript) image of the graph,
  # and store it in a file named as per the parameter.
  #--
  # Ruby-graphviz has trouble with filepaths with spaces in them,
  # so add double-quotes around the filepath/name.
  #++
  def eps(outfilename)
    @graph.output(:eps => "#{outfilename}.eps".dq)
  end

  def get_gv
    return @graph_string
  end
end 

