# Author: Bruce Tesar

require 'test/unit'
require 'shoulda'
require_relative 'quick_erc'
require_relative '../lib/constraint'
require_relative '../lib/erc'

class Test_erc < Test::Unit::TestCase
  include QuickErc # for testing further down

  def setup
    constraints = ["C1", "C2", "C3"]
    @erc1 = Erc.new(constraints, "erc1")
    @erc1.set_w("C1"); @erc1.set_l("C2"); @erc1.set_e("C3")
    @erc2 = Erc.new(constraints, "erc2")
    @erc2.set_l("C1"); @erc2.set_e("C2"); @erc2.set_w("C3")
    @erc3 = Erc.new(constraints, "erc3")
    @erc3.set_l("C1"); @erc3.set_l("C2"); @erc3.set_e("C3")
  end
  
  def teardown
    #teardown
  end

  def test_pref_queries
    assert(@erc1.w?("C1"), "C1 should prefer the winner.")
    assert(!@erc1.l?("C1"), "C1 should not prefer the loser.")
    assert(!@erc1.e?("C1"), "C1 should not be e.")
    
    assert(!@erc1.w?("C2"), "C2 should not prefer the winner.")
    assert(@erc1.l?("C2"), "C2 should prefer the loser.")
    assert(!@erc1.e?("C2"), "C2 should not be e.")

    assert(!@erc1.w?("C3"), "C3 should not prefer the winner.")
    assert(!@erc1.l?("C3"), "C3 should not prefer the loser.")
    assert(@erc1.e?("C3"), "C3 should be e.")
  end
  
  def test_constraints
    assert_equal(["C1", "C2", "C3"], @erc1.constraint_list)
  end
  
  def test_label
    erc_nl = Erc.new(["C1", "C2", "C3"])
    erc_nl.set_w("C1")
    erc_nl.set_l("C2")
    erc_nl.set_e("C3")
    assert_equal("NoLabel", erc_nl.label)
    assert_equal("NoLabel C1:W C2:L C3:e", erc_nl.to_s)
    erc_nl.label = "erc_nl"
    assert_equal("erc_nl", erc_nl.label)
    assert_equal("erc_nl C1:W C2:L C3:e", erc_nl.to_s)
  end
  
  # Copy objects, with shared label/constraints, distinct preferences
  def test_dep
    e1_dup = @erc1.dup
    e1_dup.label = "e1_dup"
    e1_dup.set_l("C3")
    assert(@erc1.e?("C3"))
    assert(e1_dup.l?("C3"))
    assert_equal("erc1", @erc1.label)
    assert_equal("e1_dup", e1_dup.label)
    # The constraint list should be the same object for both ercs
    assert(@erc1.constraint_list.equal?(e1_dup.constraint_list))
  end
  
  def test_to_s
    assert_equal("W", @erc1.pref_to_s("C1"))
    assert_not_equal("L", @erc1.pref_to_s("C1"))
    assert_equal("e", @erc1.pref_to_s("C3"))
    assert_equal("erc1 C1:W C2:L C3:e", @erc1.to_s)
    assert_equal("C1:W C2:L C3:e", @erc1.prefs_to_s)
  end
  
  def test_eql
    assert(@erc1.eql?(@erc1), "Any erc should be eql to itself.")
    assert(@erc1==@erc1, "The == op should be synonymous with eql.")
    assert(!@erc1.eql?(@erc2), "The ercs should NOT be eql.")
    assert(@erc1!=@erc2, "The == op should be synonymous with eql.")
    assert(!@erc3.eql?(@erc1), "The ercs should NOT be eql.")
    assert(@erc3!=@erc1, "The == op should be synonymous with eql.")
    assert(!@erc1.eql?(@erc3), "The ercs should NOT be eql.")
    assert(@erc1!=@erc3, "The == op should be synonymous with eql.")
    assert(!@erc3.eql?(@erc2), "The ercs should NOT be eql.")
    assert(@erc3!=@erc2, "The == op should be synonymous with eql.")
    #
    erc1a = @erc1.dup
    erc1a.label = "erc1a"
    assert("erc1", @erc1.label)
    assert(@erc1.eql?(erc1a))
    assert(@erc1==erc1a)
    assert(!(erc1a!=@erc1))
    assert(!@erc1.equal?(erc1a))
  end
  
  def test_set
    con1 = Constraint.new("C1",1)
    con1a = Constraint.new("C1",2)
    con2 = Constraint.new("C2",2)
    s = Set.new
    assert(s.add?(con1))
    assert_nil(s.add?(con1a))
    assert_nil(s.add?(con1))
    assert(s.add?(con2))
    #
    s.clear
    assert(s.add?(@erc1))
    assert_nil(s.add?(@erc1))
    erc10 = @erc1.dup; erc10.label = "erc10"
    assert_nil(s.add?(erc10))
    assert(s.add?(@erc2))
    s10 = Set.new([erc10])
    assert(s10.subset?(s))
    assert([@erc2.dup].to_set.subset?(s))
    assert(![@erc2].to_set.subset?(s10))
  end
  
  def test_set_eql
    con1 = Constraint.new("C1",1)
    con2 = Constraint.new("C2",2)
    con3 = Constraint.new("C3",3)
    con4 = Constraint.new("C4",4)
    cons = [con1, con2, con3, con4]
    #
    erc1234 = Erc.new(cons, "erc1234")
    erc1234.set_w(con1);erc1234.set_l(con2);erc1234.set_w(con3);erc1234.set_e(con4)
    erc1324 = Erc.new(cons, "erc1324")
    erc1324.set_w(con1);erc1324.set_w(con3);erc1324.set_l(con2);erc1324.set_e(con4)
    assert(erc1234.eql?(erc1324))
    assert(erc1324.eql?(erc1234))
    #
    con10 = Constraint.new("C1",10)
    con20 = Constraint.new("C2",20)
    con30 = Constraint.new("C3",30)
    con40 = Constraint.new("C4",40)
    cons = [con10, con20, con30, con40]
    assert_equal(con1, con10)
    erc10203040 = Erc.new(cons, "erc10203040")
    erc10203040.set_w(con10);erc10203040.set_l(con20);erc10203040.set_w(con30);erc10203040.set_e(con40)
    assert(erc1234.eql?(erc10203040))
  end
  
  def test_hash
    con1 = Constraint.new("C1",1)
    con2 = Constraint.new("C2",2)
    con3 = Constraint.new("C3",3)
    con4 = Constraint.new("C4",4)
    cons = [con1, con2, con3, con4]
    erc1234 = Erc.new(cons, "erc1234")
    erc1234.set_w(con1);erc1234.set_l(con2);erc1234.set_w(con3);erc1234.set_e(con4)
    erc1324 = Erc.new(cons, "erc1324")
    erc1324.set_w(con1);erc1324.set_w(con3);erc1324.set_l(con2);erc1324.set_e(con4)
    assert_equal(erc1234.hash,erc1324.hash)
    erc123 = Erc.new(cons, "erc123")
    erc123.set_w(con1);erc123.set_l(con2);erc123.set_w(con3)
    assert_equal(erc1234.hash,erc123.hash)
  end

  # Tests for conjunctive expansion of single ERCs

  context "A one-L erc" do
    setup do
      @erc = quick_erc([W,L,E,W,W,E], "oneL")
    end
    should "return a list with that same single erc" do
      expansion = @erc.conjunctive_expansion
      assert_equal(1,expansion.size)
      assert_equal(@erc,expansion.first)
    end
  end

  context "A no-L erc" do
    setup do
      @erc = quick_erc([W,E,E,W,W,E], "noL")
    end
    should "return a list with that same single erc" do
      expansion = @erc.conjunctive_expansion
      assert_equal(1,expansion.size)
      assert_equal(@erc,expansion.first)
    end
  end

  context "A two-L erc" do
    setup do
      @erc = quick_erc([W,L,L,W,W,E], "twoL")
    end
    should "return two one-L ercs" do
      expansion = @erc.conjunctive_expansion
      assert_equal(2,expansion.size)
      # Use #find, because you don't know the order the expansion ercs.
      assert(expansion.find{|e| e.eql?(quick_erc([W,L,E,W,W,E]))}, "Expansion should contain [W,L,E,W,W,E]")
      assert(expansion.find{|e| e.eql?(quick_erc([W,E,L,W,W,E]))}, "Expansion should contain [W,E,L,W,W,E]")
    end
  end

  context "A three-L erc" do
    setup do
      @erc = quick_erc([L,E,L,W,L], "threeL")
    end
    should "return three one-L ercs" do
      expansion = @erc.conjunctive_expansion
      assert_equal(3,expansion.size)
      # Use #find, because you don't know the order the expansion ercs.
      assert(expansion.find{|e| e.eql?(quick_erc([L,E,E,W,E]))}, "Expansion should contain [L,E,E,W,E]")
      assert(expansion.find{|e| e.eql?(quick_erc([E,E,L,W,E]))}, "Expansion should contain [E,E,L,W,E]")
      assert(expansion.find{|e| e.eql?(quick_erc([E,E,E,W,L]))}, "Expansion should contain [E,E,E,W,L]")
    end
  end

  # Tests for conjunctive expansion of erc lists.
  
  context "Two one-L ercs" do
    setup do
      @erc1 = quick_erc([L,E,W,W,E])
      @erc2 = quick_erc([W,E,L,W,W])
      @erc_list = [@erc1,@erc2]
    end
    should "return the same two ercs" do
      expansion = Erc.conj_expand_list(@erc_list)
      assert_equal(2, expansion.size)
      assert(expansion.find{|e| e.eql?(@erc1)}, "Expansion should contain #{@erc1.to_s}")
      assert(expansion.find{|e| e.eql?(@erc2)}, "Expansion should contain #{@erc2.to_s}")
    end
    should "group into two L-groups, each with one erc" do
      expansion = Erc.conj_expand_list(@erc_list)
      groups = expansion.group_by{|erc| erc.l_cons.first}
      assert_equal [@erc1], groups[C1]
      assert_nil groups[C2]
      assert_equal [@erc2], groups[C3]
    end
    context "and a two-L erc" do
      setup do
        @erc_list << quick_erc([E,L,L,E,W])
        @erc3a = quick_erc([E,L,E,E,W])
        @erc3b = quick_erc([E,E,L,E,W])
      end
      should "return the first two ercs and the expansion of the last into two" do
        expansion = Erc.conj_expand_list(@erc_list)
        assert_equal(4, expansion.size)
        assert(expansion.find{|e| e.eql?(@erc3a)}, "Expansion should contain [E,L,E,E,W]")
        assert(expansion.find{|e| e.eql?(@erc3b)}, "Expansion should contain [E,E,L,E,W]")
      end
      should "group into three L-groups, with C3 having two ercs" do
        expansion = Erc.conj_expand_list(@erc_list)
        groups = expansion.group_by{|erc| erc.l_cons.first}
        assert_equal [@erc1], groups[C1]
        assert_equal [@erc3a], groups[C2]
        assert_equal [@erc2, @erc3b], groups[C3]
      end
      context "and a following one-L erc" do
        setup do
          @erc4 = quick_erc([E,W,E,W,L])
          @erc_list << @erc4
        end
        should "return five ercs" do
          expansion = Erc.conj_expand_list(@erc_list)
          assert_equal(5, expansion.size)
          assert(expansion.find{|e| e.eql?(@erc1)}, "Expansion should contain #{@erc1.to_s}")
          assert(expansion.find{|e| e.eql?(@erc2)}, "Expansion should contain #{@erc2.to_s}")
          assert(expansion.find{|e| e.eql?(quick_erc([E,L,E,E,W]))}, "Expansion should contain [E,L,E,E,W]")
          assert(expansion.find{|e| e.eql?(quick_erc([E,E,L,E,W]))}, "Expansion should contain [E,E,L,E,W]")
          assert(expansion.find{|e| e.eql?(quick_erc([E,W,E,W,L]))}, "Expansion should contain [E,W,E,W,L]")
        end
        should "group into four L-groups" do
          expansion = Erc.conj_expand_list(@erc_list)
          groups = expansion.group_by{|erc| erc.l_cons.first}
          assert_equal [@erc1], groups[C1]
          assert_equal [@erc3a], groups[C2]
          assert_equal [@erc2, @erc3b], groups[C3]
          assert_nil groups[C4]
          assert_equal [@erc4], groups[C5]
        end
      end
    end
  end

  context "A list with one all-L erc" do
    setup do
      @erc_list = [quick_erc([L,L,L])]
    end
    should "return a one-L erc for each constraint" do
      expansion = Erc.conj_expand_list(@erc_list)
      assert_equal(3, expansion.size)
      assert(expansion.find{|e| e.eql?(quick_erc([L,E,E]))}, "Expansion should contain [L,E,E]")
      assert(expansion.find{|e| e.eql?(quick_erc([E,L,E]))}, "Expansion should contain [E,L,E]")
      assert(expansion.find{|e| e.eql?(quick_erc([E,E,L]))}, "Expansion should contain [E,E,L]")
    end
    should "return three L-groups" do
      expansion = Erc.conj_expand_list(@erc_list)
      groups = expansion.group_by{|erc| erc.l_cons.first}
      assert_equal [quick_erc([L,E,E])], groups[C1]
      assert_equal [quick_erc([E,L,E])], groups[C2]
      assert_equal [quick_erc([E,E,L])], groups[C3]
    end
  end

end # Class test_erc
