# Author: Bruce Tesar
# 

require 'test/unit'
require 'shoulda'
require_relative 'quick_erc'
require_relative '../lib/w_cover'

class Test_w_cover < Test::Unit::TestCase

  include QuickErc
  
  context "When no constraints have been added" do
    setup do
      @wcov = W_cover.new
    end
    should "return empty lists of w-constraints and reduced ercs" do
      assert(@wcov.w_constraints.empty?)
      assert(@wcov.reduced_ercs.empty?)
    end
    should "raise RuntimeException for L-constraint or constraint list" do
      assert_raise(RuntimeError){@wcov.l_constraint}
      assert_raise(RuntimeError){@wcov.constraint_list}
    end
  end

  context "After adding [C1,C3] covering C2" do
    setup do
      @wcov = W_cover.new
      @erc1 = quick_erc([W,L,E,W])
      @erc2 = quick_erc([E,L,W,E])
      @erc3 = quick_erc([W,L,W,E])
      @wcov.add_w_con(C1, [@erc1,@erc3])
      @wcov.add_w_con(C3, [@erc2])
      @rederc1 = quick_erc([W,L,E,E])
      @rederc2 = quick_erc([E,L,W,E])
      @wcov_dup = @wcov.dup
    end
    should "have w-cons [C1,C3] and reduced ercs [W,L,E,E],[E,L,W,E]" do
      assert_equal([C1,C3], @wcov.w_constraints)
      assert_equal([@rederc1,@rederc2], @wcov.reduced_ercs)
    end
    should "have l-constraint C2 and constraint list [C1,C2,C3,C4]" do
      assert_equal(C2, @wcov.l_constraint)
      assert_equal([C1,C2,C3,C4], @wcov.constraint_list)
    end
    should "not modify original cover when modifying a duplicate" do
      new_erc = quick_erc([E,L,E,W])
      @wcov_dup.add_w_con(C4, [new_erc])
      assert_equal([C1,C3], @wcov.w_constraints)
      assert_equal([C1,C3,C4], @wcov_dup.w_constraints)
      assert_equal([@rederc1,@rederc2], @wcov.reduced_ercs)
      assert_equal([@rederc1,@rederc2,new_erc], @wcov_dup.reduced_ercs)
    end
    should "retain equivalent L-constraint and constraint list in a duplicate" do
      new_erc = quick_erc([E,L,E,W])
      @wcov_dup.add_w_con(C4, [new_erc])
      assert_equal(@wcov.l_constraint, @wcov_dup.l_constraint)
      assert_equal(@wcov.constraint_list, @wcov_dup.constraint_list)
    end
    should "have ranking ercs [W,E,L,E],[W,E,E,L]" do
      assert_equal(quick_erc([W,E,L,E]), @wcov.ranking_ercs[0])
      assert_equal(quick_erc([W,E,E,L]), @wcov.ranking_ercs[1])
    end
    should "have a constraint list with four constraints" do
      assert_equal(4,@wcov.constraint_list.size)
      assert_equal([C1,C2,C3,C4], @wcov.constraint_list)
    end
  end

  context "after adding [W,W,L,E] and C1" do
    setup do
      @wcov = W_cover.new
      @erc1 = quick_erc([W,W,L,E])
      @wcov.add_w_con(C1, [@erc1])
      @rederc = quick_erc([W,E,L,E])
      @ranerc = quick_erc([W,L,E,E])
    end
    should "have W-constraints [C1]" do
      assert_equal([C1], @wcov.w_constraints)
    end
    should "have reduced ercs [[W,E,L,E]]" do
      assert_equal([@rederc], @wcov.reduced_ercs)
    end
    should "have ranking ercs [[W,L,E,E]]" do
      assert_equal([@ranerc], @wcov.ranking_ercs)
    end
  end

  context "after adding [W,W,L,E] and C2" do
    setup do
      @wcov = W_cover.new
      @erc1 = quick_erc([W,W,L,E])
      @wcov.add_w_con(C2, [@erc1])
      @rederc = quick_erc([E,W,L,E])
      @ranerc = quick_erc([L,W,E,E])
    end
    should "have W-constraints [C2]" do
      assert_equal([C2], @wcov.w_constraints)
    end
    should "have reduced ercs [[E,W,L,E]]" do
      assert_equal([@rederc], @wcov.reduced_ercs)
    end
    should "have ranking ercs [[L,W,E,E]]" do
      assert_equal([@ranerc], @wcov.ranking_ercs)
    end
  end

end # class Test_w_cover
