# Author: Nazarre Merchant
#

# A parser for basic syllable theory: given licit nucleuses and margins (stored in arrays)
# when parse(input) is called, an array is returned with all possible syllabifications
# with syllables boundaries delimited by '.'. Characters in 'input' that are not in 
# the @nucs & @margins array are ignored for syllabification purposes. This allows the user 
# to encode other information into the string.

# TODO: modify to allow heavy syllables. Current impl only allows one nuc segment per syll nuc.

require_relative 'parser.rb'

class BST_Parser < Parser
  
  def initialize(nucs, margins, nuc_inputs, margin_inputs)
    super(nucs, margins, nuc_inputs, margin_inputs)
    @nucs = nucs
    @margins = margins
    @nuc_inputs = nuc_inputs
    @margin_inputs = margin_inputs
  end
  
  def parse(input)
    # parse calls complete_parses which scans the input L-to-R placing possible syllable boundaries
    
    nuc = false           # Whether there is a nucleus to the right of most recent syll boundary
    pos = 0               # The location in input to the left of which the string is parsed
                          # and to the right of which parsing must occur.
    
    # Determining loc of first parseable segment
    pos+=1 while !is_valid_seg(input[pos]) and pos < input.length
    
    # no parseable segments, so don't modify input
      return [input.dup] if pos == input.length
    
    # Placing first syll (left) boundary.
    parse_in = input.dup.insert(pos, ".")
    pos+=1
    return remove_invalids(complete_parses(parse_in, nuc, pos))
  end
  
  def unparse(output)
    return desyllabify(output)
  end
  
  # Removes parses that have syllables with no nucleuses.
  # These can occur if the user passes in an input with no valid nuc, a nec. possibility.
  def remove_invalids(parses)
    
    mark_unparsed_segs = true
    parses = mark_unparsed(parses) if mark_unparsed_segs
    
    allow_unparsed = true
    return parses if allow_unparsed
    
    valid_ps = Array.new
    parses.each do |p|

      pos = 1
      # get to first syll boundary
      pos+=1 while !is_valid_seg(p[pos]) and p[pos] != '.'
      
      found_nuc = false
      remove = false
      (pos..p.length-1).each do |i|
        found_nuc = true if is_valid_nuc(p[i]) 
        remove = true if (p[i] == '.' and !found_nuc)
        found_nuc = false if (p[i] == '.' and found_nuc)  
      end
      
      valid_ps << p unless remove
    end
    return valid_ps
  end
  
  def mark_unparsed_segs(parses)
    return parses
    
    marked = Array.new
    pos = 1
    
    # get to first syll boundary
    pos+=1 while parse[pos] != '.'
    # replace all margins with Q
    
    syl_sta = pos
    syl_end = 4
    found_nuc = false
    
    
    return marked
  end
  
  def contains_unsyllabified_seg(parse)
    pos = 1
    # get to first syll boundary
    pos+=1 while !is_valid_seg(parse[pos]) and parse[pos] != '.'
    
    found_nuc = false
    remove = false
    (pos..parse.length-1).each do |i|
      found_nuc = true if is_valid_nuc(parse[i]) 
      return true if (parse[i] == '.' and !found_nuc)
      found_nuc = false if (parse[i] == '.' and found_nuc)  
    end

    return false
  end
  
  def contains_syllabified_segs_only(parse)
    return true
  end
  
  # Removes all syllable boundaries. Also, replaces each nuc and margin with its
  # corresponding nuc_input & margin_input. Finally, removes any ineffables.
  def desyllabify(parse)
    desyll = parse.dup
    desyll.tr!('.', '')
    (0..@nucs.length-1).each {|i| desyll.tr!(@nucs[i], @nuc_inputs[i])}
    (0..@margins.length-1).each {|i| desyll.tr!(@margins[i], @margin_inputs[i])}
    return remove_ineffables(desyll)
  end
  
  # Returns a string with non-margins and non-nuc symbols removed.
  def remove_ineffables(input)
    parse = input.dup
    non_p = ""
    input.each_char {|c| non_p += c if !is_valid_seg(c) && c != '.'}
    return parse.tr(non_p, '')
  end
  
  # Returns an array of the parses of 'parse'.
  # 'parse' is a string of margins, nucs, & other symbols
  # 'pos' is the position of the current parse -- to the left of pos the string 'parse'
  #       is parsed, but to the right, it is not.
  # 'nuc' is boolean, signifying whether this parse contains a syllable nucleus to
  #       the right of the rightmost syll boundary but to the left of pos.
  
  def complete_parses(parse, nuc, pos)
    # we're at end of string here, seal the deal on the final syllable, return 
    if pos == parse.length - 1
      return [parse + "."]

    # need to diff valid margin from not valid nuc: allows parsing-invisible chars to reside in-string.
    elsif (nuc and is_valid_margin(parse[pos])) or (!nuc and is_valid_nuc(parse[pos]))
      nuc_parse = parse.dup.insert(pos+1, ".")
      return complete_parses(parse, true, pos+1) if no_remaining_nucs(parse, pos)
      return complete_parses(nuc_parse, false, pos+2) if next_parseable_seg_is_nuc(parse, pos)
      return complete_parses(nuc_parse, false, pos+2).concat(complete_parses(parse, true, pos+1))
           
    # a valid nucleus has not been reached & we're not looking at a nuc so this is a margin 
    # in onset position or a non-parseable segment, so move the pos pointer one right, continue on.
    else
      return complete_parses(parse, nuc, pos+1)
    end
  end
  
  def no_remaining_nucs(parse, pos)
    @nucs.each {|n| return false if parse[pos+1..-1].scan(n).length > 0}
    return true
  end
  
  def is_valid_nuc(segment)
    return true if @nucs.include?(segment)
    return false
  end
  
  def is_valid_margin(segment)
    return true if @margins.include?(segment)
    return false
  end
  
  def is_valid_seg(segment)
    return true if @nucs.include?(segment) or @margins.include?(segment)
    return false
  end
  
  def next_parseable_seg_is_nuc(parse, pos)
    p = pos
    p+=1
    while !is_valid_seg(parse[p]) 
      p+=1
    end
    
    return true if is_valid_nuc(parse[p])
    return false
  end
end

