# Author: Nazarre Merchant
#

# A parser for basic syllable theory: given licit nucleuses and margins (stored in arrays)
# when parse(input) is called, an array is returned with all possible syllabifications
# with syllables boundaries delimited by '.'. Characters in 'input' that are not in 
# the @nucs & @margins array are ignored for syllabification purposes. This allows the user 
# to encode other information into the string.

# TODO: modify to allow heavy syllables. Current impl only allows one nuc segment per syll nuc.

require_relative 'parser.rb'

class EST1_Parser < Parser
  
  def initialize(nucs, margins, unsyl, nuc_inputs, margin_inputs, unsyl_inputs)
    super(nucs, margins, nuc_inputs, margin_inputs)
    @nucs = nucs
    @margins = margins
    @unsyl = unsyl
    @nuc_inputs = nuc_inputs
    @margin_inputs = margin_inputs
    @unsyl_inputs = unsyl_inputs
  end
  
  def parse(input)
    # parse calls complete_parses which scans the input L-to-R placing possible syllable boundaries
    
    #
    parses = Array.new
    l_parses = place_left_boundaries(input)
    l_parses.each {|p| parses.concat(place_right_boundaries(p))}
    parses.each {|p| mark_unsyllabified_segs(p)}
    
    return parses
  end
  
  def place_left_boundaries(input)
    lefts = Array.new
    lefts << input.dup
    margin_count = 0
    (0..input.length-1).each do |i|
      if @nucs.include?(input[i])
        temp_lefts = Array.new
        lefts.each do |p|
          # need an unmodified copy of p in case input[i] is preceded by a margin
          p_ons = p.dup
          temp_lefts << p.insert(i+margin_count, ".")
          temp_lefts << p_ons.insert(i+margin_count-1, ".") if @margins.include?(p_ons[i+margin_count-1]) && i+margin_count-1 > -1
        end
        margin_count += 1
        lefts = temp_lefts
      end
    end
    return lefts
  end
  
  def place_right_boundaries(input)
    rights = Array.new
    rights << input.dup
    margin_count = 0
    (0..input.length-1).each do |i|
      if @nucs.include?(input[i])
        temp_r = Array.new
        rights.each do |p|
          # need an unmodified copy of p in case input[i] is followed by a margin
          p_cod = p.dup
          temp_r << p.insert(i+margin_count+1, ".")
          temp_r << p_cod.insert(i+margin_count+2, ".") if @margins.include?(p_cod[i+margin_count+1])
        end
        margin_count += 1
        rights = temp_r
      end
    end
    rights.each {|str| str.gsub!(/\.\./, ".")}
    return rights
  end
  
  def mark_unsyllabified_segs(parse)  
    #puts "parse #{parse.to_s}"
   
    syl_b = Array.new
    (0..parse.length-1).each {|i| syl_b << i if parse[i] == "."}

    if syl_b.empty?
      replace_segs(parse, 0, parse.length-1) 
      return
    end
    
    left_b = 0
    # right_b counter:
    (0..syl_b.length-1).each do |r|
      replace_segs(parse, left_b, syl_b[r]) if !contains_nuc(parse, left_b, syl_b[r])
      left_b = syl_b[r]
    end
    replace_segs(parse, left_b, parse.length-1) if !contains_nuc(parse, left_b, parse.length-1)
  end
  
  def contains_nuc(parse, i, j) 
    (i..j).each {|s| return true if @nucs.include?(parse[s])}
    return false
  end
  
  def replace_segs(parse, i, j)
    (i..j).each do |pos|
      #if margin, replace with corresponding unsyl
      if @margins.include?(parse[pos])
        seg = parse.slice!(pos)
        seg_i = @margins.index(seg)
        parse.insert(pos, @unsyl[seg_i])
      end  
    end
  end
  
  def contains_syllabified_segs_only(parse)
    return false if parse[0] != "." || parse[parse.length-1] != "."
    
    syl_b = Array.new
    (0..parse.length-1).each {|i| syl_b << i if parse[i] == "."}

    left_b = 0
    # right_b counter:
    (1..syl_b.length-1).each do |r|
      found_nuc = false
      (left_b..syl_b[r]).each {|pos| found_nuc = true if @nucs.include?(parse[pos])}
      return false if !found_nuc
      left_b = syl_b[r]
    end

    return true
  end
  
  def unparse(output)
    return desyllabify(output)
  end

  def mark_unparsed_segs(parses)
    parses.each do |p|
      # If no syl boundary at beginning...
      # If no syl boundary at end...
      
    pos = 1
    # get to first syll boundary
    pos+=1 while !is_valid_seg(parse[pos]) and parse[pos] != '.'
    
    found_nuc = false
    remove = false
    (pos..parse.length-1).each do |i|
      found_nuc = true if is_valid_nuc(parse[i]) 
      return true if (parse[i] == '.' and !found_nuc)
      found_nuc = false if (parse[i] == '.' and found_nuc)  
    end

    return false
      
    end
    return parses
  end
  
  # Removes all syllable boundaries. Also, replaces each nuc and margin with its
  # corresponding nuc_input & margin_input. Finally, removes any ineffables.
  def desyllabify(parse)
    desyll = parse.dup
    desyll.tr!('.', '')
    (0..@nucs.length-1).each {|i| desyll.tr!(@nucs[i], @nuc_inputs[i])}
    (0..@margins.length-1).each {|i| desyll.tr!(@margins[i], @margin_inputs[i])}
    (0..@unsyl.length-1).each {|i| desyll.tr!(@unsyl[i], @unsyl_inputs[i])}   
    return remove_ineffables(desyll)
  end
  
  # Returns a string with non-margins, non-nuc, & non-unsyl symbols removed.
  def remove_ineffables(input)
    parse = input.dup
    non_p = ""
    input.each_char {|c| non_p += c if !is_valid_seg(c) && c != '.'}
    return parse.tr(non_p, '')
  end
  
  def is_valid_seg(seg)
    return true if @nucs.include?(seg) or @margins.include?(seg) or @unsyl.include?(seg)
    return false
  end

end

